// SPDX-License-Identifier: GPL-2.0
/*
 * ain-set-wake-trigger.c
 *
 * Copyright (c) 2023 Atmark Techno, Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <inttypes.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/file.h>

#define VERSION			"1.0"

#define IIO_DEVICE		"iio:device"
#define IIO_DIR			"/sys/bus/iio/devices/"
#define IIO_IN_VOLTAGE		"in_voltage"
#define IIO_RAW			"_raw"
#define IIO_SCALE		"_scale"
#define IIO_EVENTS		"events/"
#define IIO_TH_FALL		"_thresh_falling_value"
#define IIO_TH_RISE		"_thresh_rising_value"
#define IIO_TH_EITHER_EN	"_thresh_either_en"
#define OF_NODE_REG		"of_node/reg"
#define IIO_DEV_NAME		"ads1015"

#define AIN_VOL_MIN		(0U)		/* min voltage A/D conversion available in mv */
#define AIN_VOL_MAX		(5100U)		/* max voltage A/D conversion available in mv */
#define AIN_VOL_UNKNOWN		(9999999U)

#define ADC_VOLTAGE_LEN_MAX	(6)		/* max length of in_voltage0_raw + 1 */
#define ADC_SCALE_LEN_MAX	(12)		/* max length of in_voltage_scale + 1 */

#define CMD_PORT_MIN		(1)
#define CMD_PORT_MAX		(16)
#define	CMD_PORT_NOT_SET	(0xFFFF)

#define PORT_PAR_BORAD		(4)
#define BOARD_NUM_MAX		(4)

static int iio_dev_num;
static int iio_port_num;
static unsigned int cmd_board_num;
static unsigned int cmd_port_num = CMD_PORT_NOT_SET;
static unsigned int voltage_under = AIN_VOL_UNKNOWN;
static unsigned int voltage_over = AIN_VOL_UNKNOWN;

static bool trigger_enable = false;
static bool trigger_disable = false;
static bool trigger_disable_all = false;

static inline unsigned int strtouint(const char *s, char **endptr, int base)
{
	uintmax_t ret = strtoumax(s, endptr, base);

	/* same as strtoul() */
	if (ret > UINT_MAX) {
		errno = ERANGE;
		return UINT_MAX;
	}

	return (unsigned int)ret;
}

static int xstrtouint(const char *str, unsigned int *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	/* to avoid negative value */
	if (!isdigit((int)ptr[0]))
		return -1;

	errno = 0;
	*val = strtouint(ptr, &endptr, 0);

	if (errno || *endptr)
		return -1;

	return 0;
}

static int xstrtod(const char *str, double *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	errno = 0;
	*val = strtod(ptr, &endptr);

	if (errno || *endptr)
		return -1;

	return 0;
}

static double get_adc_scale()
{
	char path[MAXPATHLEN];
	char *c;
	FILE *fp;
	int  ret;
	double scale;
	char scale_str[ADC_SCALE_LEN_MAX];

	scale = 0.0;

	snprintf(path, sizeof(path), "%s%s%d/%s%d%s", IIO_DIR, IIO_DEVICE,
		 iio_dev_num, IIO_IN_VOLTAGE, iio_port_num, IIO_SCALE);
	fp = fopen(path, "r");
	if (!fp) {
		fprintf(stderr, "cannot open %s\n", path);
		return scale;
	}
	ret = fscanf(fp, "%s", scale_str);
	if (ret <= 0) {
		fprintf(stderr, "cannot read %s\n", path);
		fclose(fp);
		return scale;
	}

	fclose(fp);

	c = strchr(scale_str, '\n');
	if (c)
		*c = '\0';

	xstrtod(scale_str, &scale);
	return scale;
}

static int is_valid_voltage(unsigned int voltage)
{
	if (AIN_VOL_MAX < voltage)
		return -1;

	return 0;
}

#define IIO_MAX_NAME_LENGTH 30
static int find_iio_device_number(unsigned int board_num)
{
	const struct dirent *ent;
	FILE *fp;
	DIR *dp;
	int num;
	char path[MAXPATHLEN];
	char name[IIO_MAX_NAME_LENGTH];
	char reg[IIO_MAX_NAME_LENGTH];

	dp = opendir(IIO_DIR);
	if (dp == NULL)
		return -1;

	for (ent = readdir(dp); ent != NULL; ent = readdir(dp)) {
		if (ent->d_name[0] == '.')
			continue;
		if (strncmp(ent->d_name, IIO_DEVICE, strlen(IIO_DEVICE)))
			continue;

		sscanf(ent->d_name, IIO_DEVICE"%d", &num);
		sprintf(path, "%s%s%d/name", IIO_DIR, IIO_DEVICE, num);
		fp = fopen(path, "r");
		if (!fp)
			continue;
		fscanf(fp, "%s", name);

		fclose(fp);
		if (!strcmp(IIO_DEV_NAME, name)) {
			snprintf(path, sizeof(path), "%s%s%d/%s",
				 IIO_DIR, IIO_DEVICE, num, OF_NODE_REG);
			fp = fopen(path, "rb");
			if (!fp)
				continue;
			fread(reg, 1, 4, fp);
			fclose(fp);
			if (((board_num == 1) && (reg[3] == 0x49)) ||
			    ((board_num == 2) && (reg[3] == 0x48)) ||
			    ((board_num == 3) && (reg[3] == 0x4B)) ||
			    ((board_num == 4) && (reg[3] == 0x4A))) {
				closedir(dp);
				return num;
			} else
				continue;
		}
	}
	closedir(dp);

	return -1;
}

static bool has_multi_board(void)
{
	unsigned int i;
	bool found = false;
	for (i = 1; i <= BOARD_NUM_MAX; i++) {
		if (find_iio_device_number(i) > 0) {
			if (found)
				return true;
			else
				found = true;
		}
	}
	return false;
}

static int set_either_en(bool enable)
{
	char path[MAXPATHLEN];
	FILE *fp;
	int ret;

	snprintf(path, sizeof(path), "%s%s%d/%s%s%d%s", IIO_DIR, IIO_DEVICE,
		 iio_dev_num, IIO_EVENTS, IIO_IN_VOLTAGE, iio_port_num,
		 IIO_TH_EITHER_EN);
	fp = fopen(path, "w");
	if (!fp) {
		fprintf(stderr, "cannot open %s\n", path);
		return -1;
	}

	ret = fwrite(enable?"1":"0", sizeof(char), 1, fp);
	if (ret <= 0) {
		fprintf(stderr, "cannot write %s\n", path);
		ret = -1;
	} else
		ret = 0;

	fclose(fp);
	return ret;
}

static int set_voltage(bool is_over)
{
	char path[MAXPATHLEN];
	char vol_str[ADC_VOLTAGE_LEN_MAX];
	FILE *fp;
	int ret;
	double scale;
	unsigned int voltage;

	scale = get_adc_scale();
	if (scale == 0.0)
		return -1;

	voltage = (is_over?voltage_over:voltage_under) / scale;

	snprintf(path, sizeof(path), "%s%s%d/%s%s%d%s", IIO_DIR, IIO_DEVICE,
		 iio_dev_num, IIO_EVENTS, IIO_IN_VOLTAGE, iio_port_num,
		 is_over?IIO_TH_RISE:IIO_TH_FALL);
	snprintf(vol_str, sizeof(vol_str), "%u", voltage);
	fp = fopen(path, "w");
	if (!fp) {
		fprintf(stderr, "cannot open %s\n", path);
		return -1;
	}

	ret = fwrite(vol_str, sizeof(char),
		     strnlen(vol_str, ADC_VOLTAGE_LEN_MAX), fp);
	if (ret < 0) {
		fprintf(stderr, "cannot write %s\n", path);
		ret = -1;
	} else
		ret = 0;

	fclose(fp);
	return ret;
}

static void usage(void)
{
	fprintf(stderr, "Usage: ain-set-wake-trigger -e|-d|-a [-p port] [-o VOLTAGE] [-u VOLTAGE]\n"
		"Options:\n"
		"  -e, --enable\n"
		"      Set trigger enable.\n"
		"  -d, --disable\n"
		"      Set trigger disable\n"
		"  -a, --all-disable\n"
		"      Set trigger disable all port.\n"
		"  -p, --port\n"
		"      port number.\n"
		"  -o, --over=VOLTAGE\n"
		"      voltage is equal to or over the VOLTAGE[mV].\n"
		"  -u, --under=VOLTAGE\n"
		"      voltage is equal to or under the VOLTAGE[mV].\n"
		"VOLTAGE: Range: %u - %u\n\n"
		, AIN_VOL_MIN, AIN_VOL_MAX);
}

static inline void version(void)
{
	fprintf(stdout, "ain-set-wake-trigger %s\n", VERSION);
}

static int parse_args(int argc, char *argv[])
{
	struct option longopts[] = {
		{"port", required_argument, NULL, 'p' },
		{"over", required_argument, NULL, 'o' },
		{"under", required_argument, NULL, 'u' },
		{"enable", no_argument, NULL, 'e' },
		{"disable", no_argument , NULL, 'd' },
		{"disable-all", no_argument , NULL, 'a' },
		{"help", no_argument , NULL, 'h' },
		{"version", no_argument , NULL, 'v' },
		{0,0,0,0},
	};
	int c;

	while ((c = getopt_long(argc, argv, "+p:o:u:edahv",longopts, NULL)) != EOF) {
		switch (c) {
		case 'p':
			if (cmd_port_num != CMD_PORT_NOT_SET)
				return -1;
			if (xstrtouint(optarg, &cmd_port_num))
				return -1;
			if (cmd_port_num < CMD_PORT_MIN ||
			    cmd_port_num > CMD_PORT_MAX)
				return -1;
			cmd_board_num = (cmd_port_num / PORT_PAR_BORAD);
			cmd_board_num += (cmd_port_num % PORT_PAR_BORAD)?1:0;
			iio_port_num = (cmd_port_num-1)%PORT_PAR_BORAD;
			break;

		case 'o':
			if (xstrtouint(optarg, &voltage_over))
				return -1;
			if (is_valid_voltage(voltage_over))
				return -1;
			break;
		case 'u':
			if (xstrtouint(optarg, &voltage_under))
				return -1;
			if (is_valid_voltage(voltage_under))
				return -1;
			break;
		case 'e':
			if (trigger_enable || trigger_disable ||
			    trigger_disable_all) {
				return -1;
			}
			trigger_enable = true;
			break;
		case 'd':
			if (trigger_enable || trigger_disable ||
			    trigger_disable_all) {
				return -1;
			}
			trigger_disable = true;
			break;
		case 'a':
			if (trigger_enable || trigger_disable ||
			    trigger_disable_all) {
				return -1;
			}
			trigger_disable_all = true;
			return 0;
		case 'h':
			usage();
			return 1;
		case 'v':
			version();
			return 1;
		default:
			return -1;
		}
	}

	if (!trigger_enable && !trigger_disable && !trigger_disable_all)
		return -1;

	if (trigger_enable &&
	    (((voltage_under == AIN_VOL_UNKNOWN) &&
	     (voltage_over == AIN_VOL_UNKNOWN)) ||
	    (cmd_port_num == CMD_PORT_NOT_SET))) {
		fprintf(stderr, "enable is need port and over or/and under voltage.\n");
		return -1;
	}

	if (trigger_disable && (cmd_port_num == CMD_PORT_NOT_SET)) {
		fprintf(stderr, "disable is need port.\n");
		return -1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	int ret;
	unsigned int board_num;

	ret = parse_args(argc, argv);
	if (ret < 0) {
		usage();
		return EXIT_FAILURE;
	}
	if (ret) {
		return EXIT_SUCCESS;
	}

	if (trigger_disable_all) {
		for ( board_num = 1; board_num <= BOARD_NUM_MAX; board_num++) {
			iio_dev_num = find_iio_device_number(board_num);
			if (iio_dev_num < 0) {
				if (board_num == 1)
					fprintf(stderr, "no ain board.\n");
				break;
			}
			for (iio_port_num = 0; iio_port_num <  PORT_PAR_BORAD;
								iio_port_num++)
				set_either_en(false);
		}
	} else {
		iio_dev_num = find_iio_device_number(cmd_board_num);
		if (iio_dev_num < 0) {
			fprintf(stderr, "cannot find %u\n", cmd_board_num);
			return EXIT_FAILURE;
		}
		if (trigger_enable) {
			if (has_multi_board()) {
				fprintf(stderr, "cannot enable with multi board.\n");
				return EXIT_FAILURE;
			}

			if (set_either_en(false))
				return EXIT_FAILURE;
			if (voltage_under != AIN_VOL_UNKNOWN) {
				if (set_voltage(false))
					return EXIT_FAILURE;
			}
			if (voltage_over != AIN_VOL_UNKNOWN) {
				if (set_voltage(true))
					return EXIT_FAILURE;
			}
			if (set_either_en(true))
				return EXIT_FAILURE;
		} else {
			if (set_either_en(false))
				return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}
